﻿Imports System.Data.SqlClient
Imports System.Data
Imports System.Data.Common
Public Class SqlPersistance
    Private _Con As SqlConnection
    Private _Adapters As Dictionary(Of DataTable, SqlDataAdapter)
    Private _RankedTables As List(Of DataTable)
    Private _Dts As DataSet
    Public Sub New(connectionString As String, dts As DataSet)
        _Dts = dts
        _Con = New SqlConnection(connectionString)
        Dim sorter = New TopologicSort(dts.Tables.Cast(Of DataTable), True)
        _RankedTables = sorter.GetRankedTables
        _Adapters = New Dictionary(Of DataTable, SqlDataAdapter)
        For Each tb In _RankedTables
            Dim adp = New SqlDataAdapter("Select * from [" & tb.TableName & "]", _Con)
            Dim cmb = New SqlCommandBuilder(adp)
            _Adapters.Add(tb, adp)
        Next
    End Sub
    Public Sub FillAll()
        _Dts.Clear()
        _Con.Open()
        For Each tb In _RankedTables
            _Adapters(tb).Fill(tb)
        Next
        _Con.Close()
    End Sub
    ''' <summary> since SELECT and FROM-Clause are predefined by the datatable-structure sqlAfterFrom can contain any other valid Sql-clauses, especially WHERE. To prevent Sql-Injection-attacs it is strongly recommended to use '?' as Parameter-PlaceHolder and submit appropriate args</summary>
    Public Sub CustomFill(table As DataTable, sqlAfterFrom As String, ParamArray args() As Object)
        Dim childrenFinder = New TopologicSort(_RankedTables, False)
        For Each tb In childrenFinder.GetRankedTables(table) : tb.Clear() : Next
        Using adp = DirectCast(DirectCast(_Adapters(table), ICloneable).Clone, SqlDataAdapter)
            Dim cmd = adp.SelectCommand
            Dim splits = sqlAfterFrom.Split("?"c)
            For i = 0 To args.Length - 1
                Dim name = "@p" & i
                cmd.Parameters.AddWithValue(name, args(i))
                splits(i) = String.Concat(splits(i), " ", name, " ")
            Next
            cmd.CommandText &= " " & String.Concat(splits)
            _Con.Open()
            adp.Fill(table)
            _Con.Close()
        End Using
    End Sub
    Public Sub Save()
        _Con.Open()
        For Each tb In _RankedTables
            Dim rows = tb.Select("", "", DataViewRowState.Added Or DataViewRowState.ModifiedCurrent)
            _Adapters(tb).Update(rows) ' first send Inserts and Updates
        Next
        Dim skipSubOrderedRowsConfig = New _AcceptruleCascadeConfig(_Dts)
        For Each tb In _RankedTables
            Dim n = _Adapters(tb).Update(tb) ' send the remaining Deletes, skipping Cascade-Deleted-Rows, which will be deleted by Db itself
        Next
        skipSubOrderedRowsConfig.Restore()
        _Con.Close()
    End Sub
    Private Class _AcceptruleCascadeConfig : Inherits List(Of Tuple(Of ForeignKeyConstraint, AcceptRejectRule))
        ' AccepRule.Cascade removes subordered rows from datasets change-tracking, during Updating Database
        ' Required while Updating deleted rows, since the db deletes subordered rows already by itself.
        Public Sub New(ByVal dts As DataSet)
            For Each rl As DataRelation In dts.Relations
                Dim ck = rl.ChildKeyConstraint
                If ck Is Nothing Then Continue For
                MyBase.Add(Tuple.Create(ck, ck.AcceptRejectRule))
                ck.AcceptRejectRule = AcceptRejectRule.Cascade
            Next
        End Sub
        Public Sub Restore()
            For Each tpl In Me
                tpl.Item1.AcceptRejectRule = tpl.Item2
            Next
        End Sub
    End Class
    Private Class TopologicSort
        Private _RankedTables As List(Of DataTable)
        Private _Hsh As HashSet(Of DataTable)
        Private _ConsiderParentRelations As Boolean ' (otherwise consider ChildRelations)
        Private _Tables As IEnumerable(Of DataTable)
        ''' <summary> if considerParentRelations = False it consideres ChildRelations </summary>
        Public Sub New(tables As IEnumerable(Of DataTable), considerParentRelations As Boolean)
            _Tables = tables : _ConsiderParentRelations = considerParentRelations
        End Sub
        ''' <summary> returns all to root related tables (either parents or childs), including root as last element. If no root specified returns all tables in topologic order </summary>
        Public Function GetRankedTables(Optional root As DataTable = Nothing) As List(Of DataTable)
            _Hsh = New HashSet(Of DataTable)(_Tables)
            _RankedTables = New List(Of DataTable)(_Hsh.Count)
            If root Is Nothing Then
                While _Hsh.Count > 0 : RecurseRelatedTables(_Hsh.First) : End While
            Else
                RecurseRelatedTables(root)
            End If
            Return _RankedTables
        End Function
        ''' <summary> before adding an element to result-list, recursively add its precusers </summary>
        Private Sub RecurseRelatedTables(tb As DataTable)
            If Not _Hsh.Remove(tb) Then Return 'prevent run in Cycles
            If _ConsiderParentRelations Then
                For Each rl As DataRelation In tb.ParentRelations
                    RecurseRelatedTables(rl.ParentTable)
                Next
            Else
                For Each rl As DataRelation In tb.ChildRelations
                    RecurseRelatedTables(rl.ChildTable)
                Next
            End If
            _RankedTables.Add(tb)
        End Sub
    End Class
End Class